import { IFPCMessage } from "../models/messages/IFPCMessage";
import { IAddNodeMessage } from "../models/messages/IAddNodeMessage";
import { IRemoveNodeMessage } from "../models/messages/IRemoveNodeMessage";
import { IConnectNodesMessage } from "../models/messages/IConnectNodesMessage";
import { IDisconnectNodesMessage } from "../models/messages/IDisconnectNodesMessage";

export enum WSMsgType {
    ping = 0,
    fpc = 1,
    addNode = 2,
    removeNode = 3,
    connectNodes = 4,
    disconnectNodes = 5
}

export interface WSMessage {
    type: WSMsgType;
    data: unknown;
}

type DataHandler<T> = (data: T) => void;

const handlers: { [id in WSMsgType]?: DataHandler<unknown> } = {};

export function registerHandler(msgTypeID: WSMsgType.fpc, handler: DataHandler<IFPCMessage>);
export function registerHandler(msgTypeID: WSMsgType.addNode, handler: DataHandler<IAddNodeMessage>);
export function registerHandler(msgTypeID: WSMsgType.removeNode, handler: DataHandler<IRemoveNodeMessage>);
export function registerHandler(msgTypeID: WSMsgType.connectNodes, handler: DataHandler<IConnectNodesMessage>);
export function registerHandler(msgTypeID: WSMsgType.disconnectNodes, handler: DataHandler<IDisconnectNodesMessage>);
export function registerHandler<T>(msgTypeID: number, handler: DataHandler<T>): void {
    handlers[msgTypeID] = handler;
}

export function unregisterHandler(msgTypeID: number): void {
    delete handlers[msgTypeID];
}

export function connectWebSocket(
    path: string,
    onOpen: () => void,
    onClose: () => void,
    onError: () => void): void {
    const loc = window.location;
    let uri = "ws:";

    if (loc.protocol === "https:") {
        uri = "wss:";
    }
    uri += "//" + loc.host + path;

    const ws = new WebSocket(uri);

    ws.onopen = onOpen;
    ws.onclose = onClose;
    ws.onerror = onError;

    ws.onmessage = (e) => {
        const msg: WSMessage = JSON.parse(e.data) as WSMessage;
        // Just a ping, do nothing
        if (msg.type === WSMsgType.ping) {
            return;
        }
        const handler = handlers[msg.type];
        if (handler) {
            handler(msg.data);
        }
    };
}
